#include "raw.h"

/* 
   Takes linkage file data and recodes alleles, also calculates its
   own set of allele frequencies. 
   rawx is the x chromosome compliant equivalent.
   
   $Id: raw.c,v 2.14 2002/09/06 07:16:28 wilkinson Exp $
   
   $Log: raw.c,v $
   Revision 2.14  2002/09/06 07:16:28  wilkinson
   updated code to allow for incorporation of existing allele frequencies

   Revision 2.13  2002/09/06 03:59:02  wilkinson
   updated rebinning function to handle case where we have a small family and
   we wish to use rebinned alleles from a seperate study, where there are more
   alleles at a given marker than is in the family. The next step is to use
   the allele freq info from the larger study in a .dat file.

   Revision 2.12  2002/07/18 01:32:46  wilkinson
   modified reading of marker and distance (recomb frac) files to ignore
   lines beginning with a '#'. When reading distances, it will store the
   distance after the # and add that to the the distance to the next marker.

   Revision 2.11  2001/12/10 01:19:02  wilkinson
   added an option to produce marker files in .idx format for use in
   McPeek and Sun's prest program for detecting misspecified relationships
   (AmJHG 66:1076-1094)

   Revision 2.10  2001/11/16 01:24:36  wilkinson
   updated documentation

   Revision 2.9  2001/11/16 00:27:14  wilkinson
   Enabled the use of an input file in placed of command line arguments.
   The file template file inputs lists the general style required.

   Revision 2.8  2001/11/14 04:33:45  wilkinson
   add warnings if unable to open files for reading or writing

   Revision 2.7  2001/11/09 01:30:22  wilkinson
   added a debug flag on the command line
   added a check to ensure that zeroed frequencies (ie present in affected or
   sibs but not in founders) are set to a value of MINFREQUENCY

   Revision 2.6  2001/11/08 06:08:20  bahlo

   Put in a new option -ef which puts equal allele frequencies in the .dat
   file rather than using the unaffected founders (default)

   Revision 2.5  2001/11/06 07:34:18  bahlo
   *** empty log message ***

   Revision 2.4  2001/10/31 07:02:12  wilkinson
   Added a option -ignore to command line.
   This forces the program to keep going once it has discoverd there is no
   data on a marker listed in the marker file.  It will ignore this marker in
   the .pre and .dat files. The distance to the next marker, from the
   previous marker is just the sum of the two distances (j-1->j + j->j+1)


   Revision 2.0  2001/10/08 02:31:31  wilkinson
   Fixed bug with reading in of marker files (reading in one less than
   the desired number).
   Made getline function more robust to allow multiple spaces or tabs to
   be handled (originally all field could only be seperated by a single space
   or tab).  getline also now handles whitespace after the field being read
   Program also handles extra blank lines at the end of input files


*/

int NoAlleles[MAXLOCI];      /*number of alleles at each locus*/
int IND;                     /* actual # of lines in .ped file*/
int L;	                     /* # of marker loci*/
int liab;                    /* number of liability classes */

double freq[MAXLOCI][MAXALL];/* frequencies for all alleles at each locus */
double theta[MAXLOCI];     /*recombination fractions*/

PEDIGREES *ped;

/* default options */
option_block  op_ = {
  TRUE,			/* relabel */
  FALSE, NULL,		/* rebin */
  FALSE, NULL, NULL,	/* ABIinput */
  FALSE,		/* translateIDs */
  FALSE, NULL,	       	/* Raw Alleles */
  FALSE,		/* don't use equal allele freqs */
  FALSE,		/* sex linked */	
  FALSE,		/* ignore missing markers */
  FALSE,		/* debugging is turned off */
  FALSE, NULL,		/* idx file for McPeek and Sun's prest program */
  NULL,			/* filename for founder frequencies */
  FALSE                 /* don't use existing allele freqs */
};

/* this is the table that contains the strings used to identify
   command line options. The first number is the code used in the
   case statement for parsing the command line options. The second
   is the full form of the argument.
*/
static struct {
  int   opcode;
  char *option;     /* command line options */
  char *longoption; /* input file version (must be lowercase) */
} oplist[] = {
  0, "n", "norelabel",
  1, "r", "rebin",
  2, "a", "abi",
  3, "i", "translated_ids",
  4, "raw", "raw",
  5, "ef", "equal",
  6, "x", "x",
  7, "ignore", "ignore",
  8, "d", "debug",
  9, "m", "misspecfied",
  10, "f", "freqfile",
  11, "rf", "readfreqfile",
  -1, ""       /* marks end of list */
};

/* To add a new option, do the following:
   
   1. Add an entry to option_block in raw.h
   2. Fill in an initial value in the op_ declaration above.
   3. Give it an opcode and a command line forms.
      This does not need to match what it is called in the option
      block. Add an entry to the case statement. 
   4. Add a printf line to the options shown when run without any options.
   5. Add a case to 'switch (opcode)' block to pick up the option 
      and act upon it.
*/




/* getstring: returns array of chars, 
   ignores both leading and trailing spaces or tabs,
   returns -1 of EOF, 0 for newline and number of chars
   on unix  \n is used to signify the end of a line
   on Mac's \r is used, and
   on PC's  \r\n is used.
   so we must handle each of these cases...
   note that I'm assuming that if we see a \r we are have data from 
   either a Mac or a PC
*/
int getstring(char DataField[],FILE *ifp) { 
  char c;
  int i=0;

  while ( (c=getc(ifp)) != '\n' && c != '\r' && c != EOF ) {
    /* ignore leading white space but break on trailing white space */
    if (i == 0 && (c =='\t' || c ==' '))  
      continue;                          
    else if (i > 0 && (c =='\t' || c ==' ')) 
      break;                             
    DataField[i]=c;
    i++;
  }
  DataField[i]='\0';
  if (op_.Debug == 3) printf("DataField=%s,i=%d\n",DataField,i);

  /* now search for any trailing spaces or tabs 
     if we read anything that isn't whitespace or EOF (eg char or digit)
     then push it back on the stream. Keep newlines and EOF for return codes
  */
  if ( c =='\t' || c ==' ' ) {
    do {
      c = getc(ifp);
    } while ( c =='\t' || c ==' ');
    if (c != EOF && c != '\n' && c != '\r') ungetc(c,ifp);  
  }

  if (c == '\r') {
    /* check if MSDOS format - ie next char is \n (or EOF)*/
    c = getc(ifp);
    if (c != EOF && c != '\n') ungetc(c,ifp);  
  }

  if (c == EOF)  return(-1);
  if (c == '\n' || c == '\r') return(0);
  return(i);
}

void *c_malloc(int bytes) {
  void *p;
  p=(void *)malloc(bytes);
  if (p==NULL) {
    fprintf(stderr,"WARNING - insufficient memory available. Exiting...\n");
    exit(1);
  }
  return p;
}

/* Get some memory for PEDIGREES*/
PEDIGREES *allocate_ped() {
  int i,j;

  ped         = (PEDIGREES*)c_malloc(sizeof(PEDIGREES));
  ped->m      = (int***)c_malloc((MAXIND)*(sizeof(int**)));
  ped->rawm   = (float***)c_malloc((MAXIND)*(sizeof(float**)));
  ped->pedids = (char**)c_malloc((MAXIND)*(sizeof(char*)));
  ped->indids = (char**)c_malloc((MAXIND)*(sizeof(char*)));
  ped->faids  = (char**)c_malloc((MAXIND)*(sizeof(char*)));
  ped->maids  = (char**)c_malloc((MAXIND)*(sizeof(char*)));
  ped->misc   = (int**)c_malloc((MAXIND)*(sizeof(int*)));

  for (i=0; i < MAXIND; i++) {
    ped->pedids[i] = (char*)c_malloc((MAXCHAR)*(sizeof(char)));
    ped->indids[i] = (char*)c_malloc((MAXCHAR)*(sizeof(char)));
    ped->faids[i]  = (char*)c_malloc((MAXCHAR)*(sizeof(char)));
    ped->maids[i]  = (char*)c_malloc((MAXCHAR)*(sizeof(char)));
    ped->m[i]      = (int**)c_malloc((MAXLOCI)*sizeof(int*));
    ped->rawm[i]   = (float**)c_malloc((MAXLOCI)*sizeof(float*));
    ped->misc[i]   = (int*)c_malloc((liab)*(sizeof(int)));
    for (j = 0; j < liab; j++) {
      ped->misc[i][j] = 0;
    }
    for (j=0; j < MAXLOCI; j++) {
      ped->m[i][j] = (int*)c_malloc((2)*sizeof(int));
      ped->m[i][j][0] = 0;
      ped->m[i][j][1] = 0;
      ped->rawm[i][j] = (float*)c_malloc((2)*sizeof(float));
      ped->rawm[i][j][0] = 0;
      ped->rawm[i][j][1] = 0;
    }
  }
  return(ped);
}


/* read in .ped file. Does not currently read in raw allele values */
int read_in_ped(char *ped_file) {
  int i,j,k;
  FILE *ifp;
  char orig[MAXCHAR];
  char temp;
  int len = 1;

  i = 0;

  if ((ifp=fopen(ped_file,"r")) == NULL) {
    fprintf(stderr,"Can't open input file %s\n",ped_file);
    exit(1);
  }

  while (len != -1) { /* len is also a flag to indicate EOF*/
    len = getstring(orig,ifp);
    if (len == -1) {
      break;
    }
    if (i >= MAXIND) {
      fprintf(stderr,"WARNING - Maximum number of individuals exceeds limit of %d\n",i);
      fprintf(stderr,"Either reduce no of individuals or modify MAXIND and recompile\n");
      fprintf(stderr,"Exiting...\n");
      exit(1);
    }

    strcpy(ped->pedids[i],orig);
    getstring(ped->indids[i],ifp);
    getstring(ped->faids[i],ifp);
    getstring(ped->maids[i],ifp);

    for (j = 0; j < liab; j++) {
      fscanf(ifp,"%d",&ped->misc[i][j]);
    }

    getstring(orig,ifp);
    ped->m[i][0][0]=atoi(orig);
    len = getstring(orig,ifp);
    ped->m[i][0][1]=atoi(orig);
    if (op_.Debug == 3) printf("m[%d][0][0]=%d,m[%d][0][1]=%d\n",
			i,ped->m[i][0][0],i,ped->m[i][0][1]);
    k=1;
    while (len > 0) {
      len=getstring(orig,ifp);
      if (len <= 0) break; /* catch new line or end of file */
      ped->m[i][k][0]=atoi(orig);
      len=getstring(orig,ifp);
      ped->m[i][k][1]=atoi(orig);		
      if (op_.Debug == 3) printf("m[%d][%d][0]=%d,m[%d][%d][1]=%d\n",
			  i,k,ped->m[i][k][0],i,k,ped->m[i][k][1]);
      k++;
    }
    i++;
  }
  L=k;
  if (op_.Debug == 2) printf("L=%d\n",L);
  return(i);
}

/* print pedigree, does not yet print out raw alleles */
void print_ped() {
  int i,j,k;
  for (i=0; i < IND; i++) {
    if (i<10) {
      printf("%s ",ped->pedids[i]);
      printf("%s ",ped->indids[i]);
      printf("%s ",ped->faids[i]);
      printf("%s ",ped->maids[i]);
      for (j=0; j < liab; j++) {
	printf("%d ",ped->misc[i][j]);
      }
      for (k=0; k < L; k++) {
	printf("%d ",ped->m[i][k][0]);
	printf("%d ",ped->m[i][k][1]);
	/*
	  printf("%f ",ped->rawm[i][k][0]);
	  printf("%f ",ped->rawm[i][k][1]);
	*/
      }
      printf("\n");
    }
  }
}

/* 
   write pedigree info to pre file, 
   allow option of binned or raw allele values 
*/
void writeprefile(char *file_name, int UseBinnedAlleles) {
  int i,j,k;
  FILE *ofp;
  

  if ((ofp=fopen(file_name,"w")) == NULL) {
    fprintf(stderr,"Can't open output ped file %s",file_name);
    exit(1);
  }


  for (i=0;i<IND;i++) {
    fprintf(ofp,"%s ",ped->pedids[i]);
    fprintf(ofp,"%s ",ped->indids[i]);
    fprintf(ofp,"%s ",ped->faids[i]);
    fprintf(ofp,"%s ",ped->maids[i]);
    for (j=0;j<liab;j++)	fprintf(ofp,"%d ",ped->misc[i][j]);
    if (UseBinnedAlleles == TRUE) {
      for (k=0; k < (L-1); k++) {
	if (NoAlleles[k] == 0) 
	  continue;
	fprintf(ofp,"%d ",ped->m[i][k][0]);
	fprintf(ofp,"%d ",ped->m[i][k][1]);
      }
      if (NoAlleles[L-1] > 0) {
	fprintf(ofp,"%d ",ped->m[i][L-1][0]);
	fprintf(ofp,"%d",ped->m[i][L-1][1]);
      }
    }
    else {
      for (k=0; k < (L-1); k++) {
	if (NoAlleles[k] == 0) 
	  continue;
	fprintf(ofp,"%.3f ",ped->rawm[i][k][0]);
	fprintf(ofp,"%.3f ",ped->rawm[i][k][1]);
      }
      if (NoAlleles[L-1] > 0) {
	fprintf(ofp,"%.3f ",ped->rawm[i][L-1][0]);
	fprintf(ofp,"%.3f",ped->rawm[i][L-1][1]);
      }
    }
    if (i<(IND-1)) fprintf(ofp,"\n");
  }
  /* Apparently we don't want an newline on the last line of the file
   * as some programs are unable to cope with this 
   * fprintf(ofp,"\n"); */
  fclose(ofp);
}

main(int argc,char *argv[]) {

  char **myargv;
  char *temp="file";
  char *InputFile[NREQUIREDFILES];
  char lowerstr[LONGSTRING];
  int InputFileCounter = 0;
  int LiabilityClass   = 0;
  int UseBinnedAlleles = TRUE;
  int uselongform = FALSE;
  int i,arg, opcode, idx;

  if (argc == 2) {
    /* first check if we have an input file rather than command line 
     * arguments. In this case we will use myargv to mimic the argv array. 
     * we copy the program name into the first element of myargv
     * we send myargv as an input to readinputfile. The pointer
     * is reallocated as each new array element is added and the final
     * adress is returned. Argv is assigned to the new address
     */
    myargv    = (char **)c_malloc( sizeof(char *) );
    myargv[0] = (char  *)c_malloc( ((strlen(argv[0])+1)*sizeof(char)) );
    strcpy(myargv[0],argv[0]);
    
    argv = readinputfile (argv[1], &argc, myargv);
    uselongform = TRUE;
    /* for (i=0; i < argc; i++) printf("argv[%d] is %s\n",i,argv[i]); */
  }

  if (argc < 6) {
    print_usage(argv[0]);
    exit(1); /* if there is insufficient arguments */
  }

  /* printf("Defaults %d %d %d %d %d %d %d %d %d\n",
   * op_.relabel,op_.rebin, op_.ABIinput, op_.translateIDs, op_.RawAlleles,
   * op_.sexlinked, op_.allelefreqs, op_.ignoreMissingMarkers, op_.Debug);
   */

  for (arg=1; arg < argc; arg++) {

    if (argv[arg][0] != '-' && !uselongform || 
	uselongform && arg <= NREQUIREDFILES +1) {

      if (InputFileCounter < NREQUIREDFILES) {
	InputFile[InputFileCounter] = 
	  (char *)c_malloc((strlen(argv[arg])+1)*sizeof(char));
	strcpy(InputFile[InputFileCounter], argv[arg]);
	InputFileCounter++;
      }
      else {
	LiabilityClass = atoi(argv[arg]);
      }
    }
    else {
      opcode = -1;

      if (uselongform) {
	/* we have input files in long form from input file */
	for ( idx=0; oplist[idx].opcode != -1; idx++) {
	  for (i=0; i < strlen(argv[arg]); i++) 
	   lowerstr[i] = tolower( argv[arg][i]); 
	  lowerstr[i] = '\0';
	  if ( strcmp( lowerstr, oplist[idx].longoption ) == 0 ) {
	    opcode = oplist[idx].opcode;
	    break;
	  }
	}     
      }
      else {
	/* search for command line arguments */
	for ( idx=0; oplist[idx].opcode != -1; idx++) {
	  if ( strcmp( &argv[arg][1], oplist[idx].option ) == 0 ) {
	    opcode = oplist[idx].opcode;
	    break;
	  }
	}     
      } 
      switch (opcode) {
      case -1: 
        fprintf(stderr,"  Warning: unknown option: %s\n",argv[arg]); 
	break;
	
      case 0: 
	op_.relabel        = FALSE;     break;
      case 1: 
	op_.rebin          = TRUE;      arg++;
	op_.rebinFile      = argv[arg]; break;
      case 2: 
	op_.ABIinput       = TRUE;      arg++;
	op_.MarkerFile     = argv[arg]; arg++;
	op_.PedigreeFile   = argv[arg]; break;
      case 3: 
	op_.translateIDs   = TRUE;      break;
      case 4: 
	op_.RawAlleles     = TRUE;      arg++;
	op_.RawAllelesFile = argv[arg]; break;
      case 5:
	op_.allelefreqs    = TRUE;	break;
      case 6:
	op_.sexlinked      = TRUE;      break;
      case 7:
	op_.ignoreMissingMarkers = TRUE;break;
      case 8:
	sscanf(argv[++arg],"%d",&op_.Debug);break;
      case 9:
	op_.idxoutput	   = TRUE;	arg++;
	op_.idxFile	   = argv[arg]; break;
      case 10:
	op_.freqFile	   = argv[++arg]; break;
      case 11:
	op_.useExistingFreqs = TRUE;    
	op_.freqFile       = argv[++arg]; break;
      default: 
        fprintf(stderr,
		"Option recognized, but no code implemented yet for: %s\n",
		argv[arg]); 
        break;
      }
    }
  }
  if (InputFileCounter < NREQUIREDFILES) {
    printf("Warning: insufficient only %d of %d required files supplied\n",
	   InputFileCounter, NREQUIREDFILES);
    printf("Exiting...\n\n");
    exit(1);
  }

  /* printf("Defaults %d %d %d %d %d %d %d %d %d\n",
   * op_.relabel,op_.rebin, op_.ABIinput, op_.translateIDs, op_.RawAlleles,
   * op_.sexlinked, op_.allelefreqs, op_.ignoreMissingMarkers, op_.Debug);
   */

  /* now check command line arguments agree */
  if (op_.RawAlleles && !op_.ABIinput) {
    fprintf(stderr,"ABI input option must be called with raw allele option\n");
    exit(1);
  }

  if (op_.translateIDs && !op_.ABIinput) {
    fprintf(stderr,"ABI input option must be called with translated ID option\n");
    exit(1);
  }


  liab = LiabilityClass + 2; /* add sex and disease status */
  ped  = allocate_ped();

  if (op_.ABIinput) {
    printf("ABI style input file\n");
    if (op_.translateIDs) printf("Sample coding in ABI file\n");
    rawabi(op_.MarkerFile, op_.PedigreeFile, InputFile[0], op_.translateIDs);
  }
  else {
    printf("Linkage style .pre input file\n");		
    IND  = read_in_ped(InputFile[0]);
    if (op_.Debug == 2) printf("L=%d\nIND=%d\n",L,IND);
  }

  if (op_.rebin) printf("Rebin\n");
  else {
    printf("No rebin\n");
    op_.rebinFile = temp;
  }

  if (op_.useExistingFreqs && op_.freqFile == NULL) {
    printf("Option to use existing frequencies chosen, but no filename provided\n");
    printf("Exiting....\n\n");
    exit(1);
  }

  rawlink(InputFile[1], InputFile[3], op_.idxFile, op_.relabel, op_.rebin, op_.rebinFile,op_.allelefreqs);

  if (op_.relabel) printf("Relabel\n");
  else             printf("No relabel\n");
  
  UseBinnedAlleles = TRUE;
  writeprefile(InputFile[2], UseBinnedAlleles);

  if (op_.RawAlleles) {
    UseBinnedAlleles = FALSE;
    writeprefile(op_.RawAllelesFile, UseBinnedAlleles);
  }

  if (op_.Debug == 2) print_ped();


  if (op_.ignoreMissingMarkers) {
    for (i=0; i< L; i++) {
      if (NoAlleles[i] == 0) {
	printf("WARNING - Marker %s (%d of %d) ignored as not present in %s\n\n",
	       mark_ord[i],i+1,L,InputFile[0]);
      }
    }
  }

  exit(0);
}

/* read an input file, and insert the fields found into argv */
char **readinputfile (char *input_file, int *myargc, char *myargv[]) {
  FILE *ifp;
  char field[LONGSTRING];
  int length, i, getstring_rc, comment_line;
  int lineno=0;

  if ((ifp=fopen(input_file,"r")) == NULL) {
    fprintf(stderr,"Can't open input file %s\n",input_file);
    exit(1);
  }

  *myargc = 1; /* keep the program name */

  /* keep reading input lines until we hit an  EOF marker */
  do {
    comment_line = FALSE;
    lineno++;
    /* process each field on a line; */
    do {
      getstring_rc = getstring(field, ifp);
      if (comment_line) continue;
      length = strlen(field);

      /* search for a comment (#) character 
       * if found set the end of string marker to this location */
      for (i=0; i< length; i++) {
	if (field[i] == '#') {
	  field[i]      = '\0';
	  length = i;
	  comment_line = TRUE;
	  break;
	}
      }
      /* push the string onto the argv array if its non zero length*/
      if (length > 0) { 
	myargv = (char **) realloc(myargv, (*myargc+1)*sizeof(char *));
	myargv[*myargc] = (char *)c_malloc( ((length+1)*sizeof(char)) );

	strcpy(myargv[*myargc],field);
	(*myargc)++;

	/* printf ("Added %s to myargv, myargc is now %d line %d\n",
	 *	myargv[*myargc-1], *myargc,lineno);
	 */
      }
    }
    while (getstring_rc > 0); /* not at end of the line yet */
  } 
  while (getstring_rc != EOF);
  close(ifp);

  //  printf("closed inputfile %s and returning to main\n",input_file);

  myargv = (char **) realloc(myargv, (*myargc+1)*sizeof(char *));
  myargv[*myargc] = NULL; /* to maintain the standard convention for argv */

  /*
   * for (i=0; i < *myargc; i++) printf("myargv[ %d ] is %s\n",i,myargv[i]);
   * printf("myargv is at %u\n",myargv);
   */

  return myargv;
}


void print_usage(char *program_name) {

  printf("Usage:\n%s inputs_file\n\n", program_name);
  printf("or\n\n");
  printf("%s rawpedfile recfile newpedfile newdatfile liability_classes [Options]\n\n", program_name);
  printf("If an input file is used, please see the default input file\n");
  printf("inputs (supplied with the distribution) for information on the\n");
  printf("required format\n\n");
  printf("rawpedfile  - linkage format .pre file with raw alleles (from pedman)\n");
  printf("              or a raw ABI file (from AGRF, needs -a option)\n");
  printf("recfile     - recombination fractions or cM file\n");
  printf("newped_file - new recoded linkage file\n");
  printf("newdat_file - newly created .dat file\n");
  printf("liability_classes = 0 - if no liability column in the pre file\n\n");
  printf("Options:\n");
  printf("-n if no relabelling is desired. Relabel is the default.\n");
  printf("-r <file> if rebinning is desired. No rebin is the default.\n");
  printf("          <file> is the file with allele mappings.\n");
  printf("-a <file1> <file2> if raw ABI file. Linkage style .pre file is the default.\n");
  printf("                   <file1> - file with marker order.\n");
  printf("                   <file2> - file with pedigree.\n\n");
  printf("-i if pedigree file for raw ABI data has translational sample ID as well.\n\n");
  printf("-m <file>  write marker information (.idx format) to <file> for use in\n");
  printf("           McPeek and Sun's prest program for detecting\n");
  printf("           misspecified relationships (AmJHG 66:1076-1094)\n\n");
  printf("-raw <file> create an additional file in .pre format containing\n");
  printf("            the raw allele values (decimal value instead of integer)\n");
  printf("            this may only be used with the -a option\n\n");
  printf("-x  use this flag if data is sex-linked (ie x chromosome)\n\n");
  printf("-ef  use equal allele frequencies rather than estimating from founders\n\n");
  printf("-f <file> write frequency data to named file (default is freqs.out)\n\n");
  printf("-rf <file>) read allele frequencies from specified file. Use this option\n");
  printf("            when the pedigree is too small to determine reliable freqency\n");
  printf("            information.  You may also want to use the rebinning option to\n");
  printf("            ensure that you are using the same alleles\n");
  printf("-ignore  ignore markers that are listed in the marker file\n\n");
  printf("         but are not present in the data. Nothing regarding these markers\n");
  printf("         will be written in the .pre or .dat files, and the map\n");
  printf("         distances between adjcant markers will be the simple linear\n");
  printf("         over the total distance (which may not be strictly true)\n\n");
  printf("-d DebugLevel  turn on debugging (see documentation for details)\n");
  printf("\n");
  printf("$Date: 2002/09/06 07:16:28 $\n");
  printf("Melanie Bahlo (bahlo@wehi.edu.au)\n");
  printf("Chris Wilkinson (wilkinson@wehi.edu.au)\n");
  printf("Not to be used without the authors' permission.\n");
  printf("See linkprep.txt for more details on options and how to use %s.\n",
	 program_name);

}
