/*
  $Id: raw.h,v 2.10 2002/09/06 07:16:28 wilkinson Exp $
  $Log: raw.h,v $
  Revision 2.10  2002/09/06 07:16:28  wilkinson
  updated code to allow for incorporation of existing allele frequencies

  Revision 2.9  2002/07/18 01:33:19  wilkinson
  modified reading of marker and distance (recomb frac) files to ignore
  lines beginning with a '#'. When reading distances, it will store the
  distance after the # and add that to the the distance to the next marker.

  Revision 2.8  2001/12/10 01:19:03  wilkinson
  added an option to produce marker files in .idx format for use in
  McPeek and Sun's prest program for detecting misspecified relationships
  (AmJHG 66:1076-1094)

  Revision 2.7  2001/11/16 00:27:15  wilkinson
  Enabled the use of an input file in placed of command line arguments.
  The file template file inputs lists the general style required.

  Revision 2.6  2001/11/09 01:30:23  wilkinson
  added a debug flag on the command line
  added a check to ensure that zeroed frequencies (ie present in affected or
  sibs but not in founders) are set to a value of MINFREQUENCY

  Revision 2.5  2001/11/08 06:08:48  bahlo

  extra option of -ef in option block

  Revision 2.4  2001/10/31 07:02:14  wilkinson
  Added a option -ignore to command line.
  This forces the program to keep going once it has discoverd there is no
  data on a marker listed in the marker file.  It will ignore this marker in
  the .pre and .dat files. The distance to the next marker, from the
  previous marker is just the sum of the two distances (j-1->j + j->j+1)

  Revision 2.2  2001/10/25 07:46:03  wilkinson
  added support for sex linked data

  Revision 2.1  2001/10/11 06:46:04  wilkinson
  file modified for addition of -raw option for saving of raw allele values

  structure also added for specifying command line options

  Revision 2.0  2001/10/08 02:31:31  wilkinson
  Fixed bug with reading in of marker files (reading in one less than
  the desired number).
  Made getline function more robust to allow multiple spaces or tabs to
  be handled (originally all field could only be seperated by a single space
  or tab).  getline also now handles whitespace after the field being read
  Program also handles extra blank lines at the end of input files

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <locale.h>

#define SIBLOCI 4	/* this has to stay the way it is*/
#define MAXLOCI 100	/* maximum # of Markers, can be reduced or increased*/
#define MAXALL 80	/* maximum # of alleles, can be reduced or increased*/
#define MAXGENO 500 /* should be approx MAXLOCI*(MAXLOCI+1)/2*/
#define FAM 1000	/* maximum number of families,can be reduced or increased*/
#define MAXCHAR 20	/*maximum size of strings for pedigree,id labels*/ 
#define MAXIND 9000	/*maximum no of lines in the ped file*/ 
#define MAXENT 9000*50
#define NREQUIREDFILES 4
#define TRUE  1
#define FALSE 0

#define LONGSTRING 255

#define MALE   1
#define FEMALE 2

#define UNAFFECTED 1
#define AFFECTED   2

#define MINFREQUENCY 1e-6
/* structure for abi data */
typedef struct{
  char **ID;           /* ind id need to link with indid */
  char **markers;      /* marker id */
  int **all_pairs;     /* allele pair data */
  float **rawall_pairs; /* raw allele pair data */
} DATA;

/* Sample structure */
typedef struct{
  int ***m;       /* m[i][j][k] ith ind,marker j, allele k */
  float ***rawm;   /* rawm[i][j][k] ith ind,marker j, raw allele k */
  char **pedids;
  /* 	pedids[i][0] is the pedid of ind i ie line i in ped file,
	indids[i][1] is indid of ind i,
	faids[i][2] is father's id of ind i, 
	maids[i][3] is mother's id of ind i,*/
  char **indids;
  char **faids;
  char **maids;
  int **misc; 
  /* misc[i][0]=ind i's sex, misc[i][1]=ind i's affection status*/
} PEDIGREES;


/* define a structure to hold options 
   
   To add a new option, do the following:
   
   1. Add an entry to option_block in raw.h
   2. Fill in an initial value in the op_ declaration in raw.c
   3. Give it an opcode and a command line forms.
      This does not need to match what it is called in the option
      block. Add an entry to the case statement. 
   4. Add a printf line to the options shown when run without any options.
   5. Add a case to 'switch (opcode)' block to pick up the option 
      and act upon it.
*/

typedef struct {
  int    relabel;
  int    rebin;
  char   *rebinFile;
  int    ABIinput;
  char   *MarkerFile;
  char   *PedigreeFile;
  int    translateIDs;
  int    RawAlleles;
  char   *RawAllelesFile;
  int    allelefreqs;
  int    sexlinked;
  int    ignoreMissingMarkers;
  int    Debug;
  int    idxoutput;
  char   *idxFile;
  char   *freqFile;
  int    useExistingFreqs;
} option_block;

/* 
#define DEBUG 0
        DEBUG >1 debug turned on, 
	2 for reading in data for -a option
	3 for readline
	5 for rawlink reorder function
	6 for rawlink freq function
	10 for endo1.pl output
	11 for rawabi.c checks
	13 for rawabi.c create .pre files
*/


/* external variables */
extern int NoAlleles[MAXLOCI];
extern int IND;
extern int L;
extern int liab;

extern char mark_ord[MAXLOCI][MAXCHAR];

extern double freq[MAXLOCI][MAXALL];
extern double theta[MAXLOCI];

extern option_block op_;
extern DATA *dat;
extern PEDIGREES *ped;

/* function prototype definitions */
int       getstring(char DataField[],FILE *ifp);
void      *c_malloc(int bytes);
PEDIGREES *allocate_ped();
int       read_in_ped(char *ped_file);
void      print_ped();
void      writeprefile(char *file_name, int UseBinnedAlleles);
char	  **readinputfile (char *input_file, int *argc, char *argv[]);
void      print_usage(char *program_name);


/* prototypes in rawlink.c */
void cm_to_fract();
double to_recombfrac(double cm);
int  check_all(int allele,int locus,int a[L][MAXALL]);
void check_a_matrix(int a[L][MAXALL]); 
void collect(int a[L][MAXALL]);
void reorder(int a[L][MAXALL]);
void zero(int a[L][MAXALL]);
void HandleMissingMarker(int marker_j);
void newbins(int bin1[L][MAXALL][2],char *rebinfile,int a[L][MAXALL]);
void print_bin(int bin1[L][MAXALL][2]);
void rebinned(int a[L][MAXALL],int bin1[L][MAXALL][2]);
void new_ped(int a[L][MAXALL]);
void frequencies(int a[L][MAXALL],int relabel);
void new_dat(char *recfile,char *datfile,int a[L][MAXALL],int relabel);
void rawlink(char *recfile, char *datfile, char *idxfile, int relabel, 
	       int rebin, char *rebinfile,int eqall);

/* function prototypes in rawabi.c */
DATA *allocate_data(int NoInd, int NoLoci);
int  missingdata(char *datastr, char *miss[], int miss_sym);
int  marker_data(char *ped_file);
int  read_in_partped(char *template_file,int abi_sample);
void create_pre(int TOT,int abi_sample);
void rawabi(char *markerfile, char *template_file, char *ped_file, 
	    int abi_sample);





